//by typing list_manifolds n" with 0<n<=9, you get a list of all irreducible 3-manifolds having
//complexity n, as a .tex file (which must be LaTex-ed thereafter)
//
//bibliography:
//
//[1] B.Martelli, C.Petronio "Complexity of geometric three-manifolds", math.GT/0303249
//
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include <fstream>
#include <list>
#include <vector>
#include <iomanip>
#include <algorithm>

#define LENS                    0
#define ELLIPTIC                1
#define FLAT                    2
#define NIL                     3
#define H2S1                    4
#define SL2                     5
#define SOL                     6
#define GRAPH                   7

#define FALSE               0
#define TRUE                1

typedef int BOOL;

#include "list_manifolds.h"

char nome_geometria[9][20] = {"lens", "other elliptic", "flat", "Nil", "H2 x S1", 
                              "S\\!L2", "Sol", "other graph", "hyperbolic"};
char nome_geometria_tex[9][30] = {"lens", "other elliptic", "flat", "Nil", 
																	"$\\mathbb{H}^2 \\times S^1$", 
																	"${\\rm SL}_2$", "Sol", "other graph", "hyperbolic"};

//some maxima for parameters, which are certainly big enough for complexity <=9
int num_fibre_max = 5;
int chi_min = 0;
int Seifert_righe_per_pagina = 35;
int graph_righe_per_pagina = 26;

//set 'TRUE' if you want the output to be a .tex file, 
//and 'FALSE' if you want a text file.
BOOL output_in_tex = TRUE;

//five important matrices. It is shown in [1] that
//every 2x2 integer matrix A with |det A| = 1 can be written
//uniquely as eS[i0]J....JS[in]S[0]^m, with e=1 or -1, i0=1,2,3, in=1,2,3, ik=1,2 for k!=0,n,
//and m=0 or 1.
//the norm |A| is then the number of J's in this unique expression.
matrice S[3] = {matrice(1,-1,0,-1), matrice(-1,0,-1,1), matrice(0,1,1,0)};
matrice J(-1,0,0,1);
matrice I(1,0,0,1);

//the norm |a,b| of a pair of coprime (a,b)
//defined as |0,1|=|1,0|=|1,1|=0, with 
//|a+b,b|=|a,a+b|=|a,b|+1,   |a,b|=|-a,-b|,   |a,-b|=|a,b|+1    for all a,b>0
int norma(int a, int b) { 
  if (a<0) return norma(-a,-b);
  if (b<0) return norma(a,-b)+1;
  if (a==1 && b==1) return 0;
  if (a==0 || b==0) return 0;
  a > b ? a=a-b : b=b-a;
  return norma(a,b)+1;
}

//the norm |A| for a matrix A with |det A|=1
int norma(const matrice& A) {
  if (A == I || A == -I) return 0;
  if (A == J || A == -J) return 1;
  matrice B(0,1,1,0);
  if (A[0][1] == 0 || A[1][1] == 0) return norma(A*B);
  if (A[1][0] == 0) return norma(B*A);
  //now the matrix is of type (ab)(cd) with b,c,d != 0
  if (A[0][0] == 0) {
    if (A[0][1]<0) return norma(-A);
    return  (A[1][0]>0 ? abs(A[1][1]) : abs(A[1][1]-1));
  }
  //now all entries in A are !=0
  std::vector<int> pos_max(2,0); 
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      if (abs(A[i][j]) > abs(A[pos_max[0]][pos_max[1]])) {
        pos_max[0] = i; pos_max[1] = j;
      }
  if (pos_max[1]) return norma(A*B);
  if (pos_max[0]) return norma(B*A);
  if (A[0][0]<0) return norma(-A); 
  //now the entries of A having maximal absolute value is A[0][0], which is also positive
  matrice C(1,0,1,1);
  matrice D(1,0,-1,1);
  return norma (A*(A[0][1]>0 ? D : C)) + 1;
}

//MCD
unsigned int MCD(int a, int b) {
  if (!a && !b) return 1;
  if (!a) return abs(b);
  if (!b) return abs(a);
  unsigned int mcd=1; a=abs(a); b=abs(b);
  for (unsigned int i=2; i<=a && i<=b; i++)
    if (!(a % i) && !(b % i)) {
      mcd*=i; a/=i; b/=i; i=1;
    }
  return mcd;
}

//default constructor for a matrix
matrice::matrice() {
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      val[i][j] = 0;
}

//constructor for a matrix via entries
matrice::matrice(int a, int b, int c, int d) {
  val[0][0] = a; val[0][1] = b;
  val[1][0] = c; val[1][1] = d;
}

//returns the sum of the absolute values of the entries
int somma_abs(const matrice& A) {
  int ris = 0;
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      ris += abs(A[i][j]);
  return ris;
}

//an ordering on matrices
BOOL operator<(const matrice& A, const matrice& B) {
  if (somma_abs(A) < somma_abs(B)) return TRUE;
  if (somma_abs(A) > somma_abs(B)) return FALSE;  
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++) {
      if (A[i][j] > B[i][j]) return TRUE;
      if (A[i][j] < B[i][j]) return FALSE;
    }
  return FALSE;
}

//matrix * matrix
matrice operator*(const matrice& A, const matrice& B) {
  matrice ris;
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      for (int k=0; k<2; k++)
        ris[i][j]+=A[i][k] * B[k][j];
  return ris;
}

//matrix - matrix
matrice operator-(const matrice& A) {
  matrice ris;
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      ris[i][j] = -A[i][j];
  return ris;
}

//matrix == matrix
BOOL operator==(const matrice& A, const matrice& B) {
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      if (A[i][j] != B[i][j]) return FALSE;
  return TRUE;
}

//det(A)
int det(const matrice& A) {
  return (A[0][0]*A[1][1] - A[0][1]*A[1][0]);
}

//inverse of A
matrice inversa(const matrice& A) {
  matrice ris(A[1][1], -A[0][1], -A[1][0], A[0][0]);
  if (det(A)<0) ris = -ris;
  return ris;
}

//change orientation of Seifert manifold
void Seifert_manifold::inverti() {
  for (int i=0; i<param.size(); i++)
    param[i][1] = param[i][0] - param[i][1];
  b = -b-param.size();
}

//returns Euler number of Seifert manifold
float Seifert_manifold::e() const {
  float ris = b;
  for (int t=0; t<param.size(); t++)
    ris += (float) param[t][1]/param[t][0];
  return ris;
}

//returns Euler characteristic of the base orbifold of the Seifert manifold
float Seifert_manifold::chi() const {
  float ris = chi_f;
  for (int t=0; t<param.size(); t++)
    ris -= (1- (float) 1/param[t][0]);
  return ris;
}

//returns the geometry of the Seifert manifold
int Seifert_manifold::geometria() const {
  BOOL e_is_zero = (e()<0.001 && e()>-0.001);
  BOOL chi_is_zero = (chi()<0.001 && chi()>-0.001);
  int ris;
  if (chi_is_zero) ris = (e_is_zero ? FLAT : NIL);
  else {
    if (!e_is_zero && chi()>0) ris = ELLIPTIC;
    if (e_is_zero && chi()<0) ris = H2S1;
    if (!e_is_zero && chi()<0) ris = SL2;
  }
  return ris;
}

//writes data of Seifert manifold on screen
void Seifert_manifold::scrivi(const int& quanti) const {
  int num_fibre = param.size();
  if (output_in_tex) {
    std::cout << " $" << quanti << "$ & $";
    for (int t=0; t<num_fibre; t++)
      std::cout << "(" << param[t][0] << ", " << param[t][1] << ") ";
    std::cout << "$ & $" << b << "$ & ";
    std::cout << nome_geometria_tex[geometria()];
    std::cout << "\\\\" << std::endl;
  }
  else {
    std::cout << std::setw(4) << quanti << ": ";
    for (int t=0; t<num_fibre; t++)
      std::cout << "(" << param[t][0] << ", " << param[t][1] << ") ";
    std::cout << "b = " << std::setw(3) << b << "   ";
    std::cout << std::setw(20) << nome_geometria[geometria()];
    std::cout << "\n";
  }
}

//auxiliary complexity function, whose minimum on representations of the
//manifold gives c3
//(recall that the same Seifert manifold can be represented via non-normalized parameters
//in many ways)
int Seifert_manifold::d3() const {
  int somma_norme = 0;
  for (int t=0; t<param.size(); t++)
    somma_norme += norma(t);
  int ris = std::max(0,b-1+chi_f) - 6*(chi_f-1) + somma_norme + 2*param.size();
  return ris;
}

//sets the data of the manifold in a standard form
BOOL Seifert_manifold::forma_standard() const {
  Seifert_manifold M = *this;
  sort(M.param.begin(), M.param.end());
  if (M < *this) return FALSE;
  M.inverti();
  sort(M.param.begin(), M.param.end());
  return (!(M < *this));
}

//an ordering on presentations of Seifert manifolds
BOOL operator<(const Seifert_manifold& M1, const Seifert_manifold& M2) {
	if (M1.chi_f < M2.chi_f) return TRUE;
	if (M1.chi_f > M2.chi_f) return FALSE;	
  if (M1.b > M2.b) return TRUE;
  if (M1.b < M2.b) return FALSE;
  if (M1.param < M2.param) return TRUE;
  if (M1.param > M2.param) return FALSE;
  return FALSE;
}

//tells if the Seifert manifold is in the set Cij (see [1])
BOOL Seifert_manifold::cij() const {
  BOOL ris = FALSE;
  if (param.size() == 3 && param[0][0] == 2 && chi_f == 2 && b==-1) {
    int prodotto = 1;
    for (int t=0; t<param.size(); t++)
      prodotto*=param[t][1];
    if (prodotto == 1) ris = TRUE;
  }
  return ris;
}

//tells if the Seifert manifold is obtained from B5 (see [1])
BOOL Seifert_manifold::from_B5() const {
  if (param.size() == 3) {
    tensor2 par = param;
    sort(par.begin(), par.end());
    if (b!=-1) return FALSE;
    if (par[0][0] == 2 && par[1][0]==3 && par[1][1]==1 && 
        par[2][1]!=1 && par[2][0] > 5*par[2][1]) return TRUE;
  }
  return FALSE;
}

//Seifert manifold == Seifert manifold
BOOL operator==(const Seifert_manifold& M1, const Seifert_manifold& M2) {
  return (M1.chi_f == M2.chi_f && M1.param == M2.param && M1.b == M2.b);
}

//graph manifold == graph manifold
BOOL operator==(const graph_manifold& M1, const graph_manifold& M2) {
  return (M1.collegamenti == M2.collegamenti && M1.A == M2.A && M1.blocchi == M2.blocchi);
}

//changes orientation of the i-th Seifert block of the graph manifold 
//(the 0-th block is the first one!)
void graph_manifold::inverti(int b) {
  blocchi[b].inverti(); blocchi[b].b = 0;

  //effect on A of changing every (p,q) into (p,-q)
  for (int i=0; i<A.size(); i++) {
    if (collegamenti[i][0] == b)
      A[i] = A[i] * J;
    if (collegamenti[i][1] == b)
      A[i] = J * A[i];
  }

  //effect on a neighborhood A[j] of changing every (p,-q) into (p,p-q)
  matrice B(1,0,-1,1);
  matrice C(1,0,1,1);
  int j = 0;
  while (collegamenti[j][0]!=b && collegamenti[j][1]!=b) j++;
  for (int i=0; i<blocchi[b].quante_fibre(); i++)
    (collegamenti[j][1] == b) ? A[j] = B * A[j] : A[j] = A[j] * C;
}

//rotate the blocks in the graph manifold, i.e. from B-B'-B'' to B''-B''-B
void graph_manifold::ruota() {
  reverse(blocchi.begin(), blocchi.end());
  reverse(A.begin(), A.end());
  reverse(collegamenti.begin(), collegamenti.end());
  for (int t=0; t<A.size(); t++)
    A[t] = inversa(A[t]);
  int n = blocchi.size();
  for (int i=0; i<collegamenti.size(); i++) {
    for (int j=0; j<=1; j++)
      collegamenti[i][j] = n-1 - collegamenti[i][j];
    reverse(collegamenti[i].begin(), collegamenti[i].end());
  }
}

//ordering on graph manifolds
BOOL operator<(const graph_manifold& M1, const graph_manifold& M2) {
  if (M1.d3() < M2.d3()) return TRUE;
  if (M1.d3() > M2.d3()) return FALSE;
  if (M1.blocchi < M2.blocchi) return TRUE;
  if (M1.blocchi > M2.blocchi) return FALSE;
  if (M1.A < M2.A) return TRUE;
  if (M1.A > M2.A) return FALSE;
  return FALSE;
}

//tells whether the i-th Seifert block is of type D(2,1)(2,1) 
//(the 0-th is the first one)
BOOL graph_manifold::due_fibrazioni(int b) const {
  return (blocchi[b].chi_f == 1 && blocchi[b].param.size() == 2 &&
					blocchi[b].param[0][0] == 2 && blocchi[b].param[1][0] == 2);
}

//tells whether there are contiguous Seifert blocks with matching fibrations
BOOL graph_manifold::is_JSJ() const {
  if (is_Sol()) return FALSE;
  for (int t=0; t<A.size(); t++) {
    matrice B = A[t];
    std::vector<int> vic = collegamenti[t];
    if (B[0][1] == 0) return FALSE;
    if (due_fibrazioni(vic[0]) && B[0][0] == B[0][1]) return FALSE;
    if (due_fibrazioni(vic[1]) && B[0][1] * B[1][1] == -1) return FALSE;
    if (due_fibrazioni(vic[0]) && due_fibrazioni(vic[1]) && 
        (B[0][0] - B[0][1]) * (B[1][0] - B[1][1]) == -1) return FALSE;
  }
  return TRUE;
}  

//tells if it is Sol
BOOL graph_manifold::is_Sol() const {
  //checks whether it fibers over an 1-orbifold as a Sol-manifold
  return (!is_Seifert() && blocchi.size() == 2 && due_fibrazioni(0) && due_fibrazioni(1));
}

//tells if it is Seifert
BOOL graph_manifold::is_Seifert() const {
  for (int t=0; t<A.size(); t++) {
    matrice B = A[t];
    BOOL fibration_extends = FALSE;
    if (B[0][1] == 0) fibration_extends = TRUE;
    if (due_fibrazioni(t) && B[0][0] == B[0][1]) fibration_extends = TRUE;
    if (due_fibrazioni(t+1) && B[0][1] * B[1][1] == -1) fibration_extends = TRUE;
    if (due_fibrazioni(t) && due_fibrazioni(t+1) && 
        (B[0][0] - B[0][1]) * (B[1][0] - B[1][1]) == -1) fibration_extends = TRUE;
    if (!fibration_extends) return FALSE;
  }
  return TRUE;
}

BOOL graph_manifold::orientable() const {
	if (line()) return TRUE;
	return (det(A[0]) == -1);
}

BOOL graph_manifold::line() const {
  for (int i=0; i<collegamenti.size(); i++)
    if (collegamenti[i][0] != i || collegamenti[i][1] != i+1)
      return FALSE;
  return TRUE;
}

BOOL graph_manifold::self_assembling() const {
	return (collegamenti[0][0] == 0 && collegamenti[0][1] == 0);
}

//dice se e' simmetrica la struttura rispetto ad una rotazione.
//(qui controlla solo che non siano 3 fibre e 2 fibre)
BOOL graph_manifold::ruotabile() const {
	if (blocchi[0].quante_fibre() == 3 ||
			blocchi[1].quante_fibre() == 3) return FALSE;
	return TRUE;
}

//dice quanti collegamenti ha un blocco
int graph_manifold::quanti_collegamenti(int num_blocco) const {
	int ris=0;
	for (int i=0; i<collegamenti.size(); i++)
		for (int j=0; j<2; j++)
			if (collegamenti[i][j] == num_blocco) ris++;
	return ris;
}

//se blocchi[num_blocco] ha due uscite, "aggiunge t alla prima e toglie param alla seconda"
void graph_manifold::cambia_sezione(int num_blocco, int param) {
	if (quanti_collegamenti(num_blocco)!=2) return;
	int gia_visti = 0;
	for (int i=0; i<collegamenti.size(); i++)
		for (int j=0; j<2; j++)
			if (collegamenti[i][j] == num_blocco) {
				if (j==0) {
					int param2 = (gia_visti ? -param : param);
					matrice L(1, 0, param2, 1);
					A[i] = A[i]*L;
				}
				if (j==1) {
					int param2 = (gia_visti ? param : -param);
					matrice L(1, 0, param2, 1);
					A[i] = L*A[i];
				}
				gia_visti++;
			}
	/*matrice L(1, 0, param, 1);
		if (num_blocco == 1) {
		A[0] = L*A[0];
		A[1] = A[1]*L;
		}*/
}

//sets the graph manifold in a standard form
//(works if the graph is a line)
void graph_manifold::set_forma_standard() {
  graph_manifold N = *this;
  const int l = N.blocchi.size();
  for (int ruota = 0; ruota<=ruotabile(); ruota++) {
    graph_manifold M = N;
    if (ruota) M.ruota();
    std::vector<BOOL> inv(l, 0);
    std::vector<BOOL> fine(l, 0);
    do {
      graph_manifold M2 = M;
      
      //inverts orientations on the k-th Seifert block if inv[k]!=0
      for (int k=0; k<l; k++)  
				if (inv[k]) M2.inverti(k);

			//if there are not 2 blocks, then there are block with two links, and more can be done
			for (int z=0; z==0 || M2.blocchi.size()!=2 && z<=2; z++) {
				int num_blocco = (M2.blocchi.size()==3 ? 1:0);
				if (z==1)
					M2.cambia_sezione(num_blocco,1);
				if (z==2)
					M2.cambia_sezione(num_blocco,-2);

				//chooses one matrix up to sign when possible
				if (!M2.self_assembling()) 
					for (int n=0; n<M2.A.size(); n++) 
						if (-M2.A[n] < M2.A[n]) M2.A[n] = -M2.A[n];
				
				//sort parameters on each block
				for (int n=0; n<l; n++) 
					sort(M2.blocchi[n].param.begin(), M2.blocchi[n].param.end());
				
				//if the new graph manifold is "smaller", then transforms into it
				if (M2 < *this) *this = M2;
			}
      inv[l-1] = 1-inv[l-1];
      for (int t=l-1; t>=1; t--)
        if (inv[t]) break;
        else inv[t-1] = 1-inv[t-1];
    } while (inv != fine);
  }
}

//tells whether the graph manifold is in a standard form
BOOL graph_manifold::forma_standard() const {
  graph_manifold N = *this;
  N.set_forma_standard();
  return (N == *this);
}

//writes data of the graph manifold on the screen
void graph_manifold::scrivi(int quanti) const {
  int max_param = 0;
  for (int i=0; i<blocchi.size(); i++)
    if (max_param < blocchi[i].quante_fibre()) max_param = blocchi[i].quante_fibre();
  if (output_in_tex) {
    std::cout << " $" << quanti << "$ ";
    for (int i=0; i<blocchi.size(); i++) {
      std::cout << "& $ ";
      for (int j=0; j<blocchi[i].quante_fibre(); j++)
				std::cout << "(" << blocchi[i].param[j][0] << ", " << blocchi[i].param[j][1] << ") ";
      std::cout << "$";
      if (i<A.size()) {
				std::cout << "& $\\matr";
				for (int t=0; t<=1; t++)
					for (int u=0; u<=1; u++)
						std::cout << "{" << A[i][t][u] << "}";
				std::cout << "$";
      }
    }
    std::cout << "\\\\" << std::endl;
  }
  else {
    for (int i=0; i<max_param ;i++) {
      (i==0) ? std::cout << std::setw(4) << quanti << ": " : std::cout << std::setw(6) << "";
      for (int j=0; j<blocchi.size(); j++) {
				if (blocchi[j].quante_fibre() > i) 
					std::cout << "(" << blocchi[j].param[i][0] << ", " << blocchi[j].param[i][1] << ")   ";
				else std::cout << std::setw(9) << "";
				if (i<2 && j < blocchi.size()-1)
					std::cout << "|" << std::setw(2) << A[j][i][0] << std::setw(3) << A[j][i][1] << "|   ";
				else std::cout << std::setw(10) << "";
      }
      std::cout << "\n";
    }
  }
}

//auxiliary complexity function, which does not depend on the matrices
int graph_manifold::d3_senza_matrici() const {
	int chi_tot = 0;
	int ris = 0;
	for (int t=0; t<blocchi.size(); t++) {
		chi_tot += blocchi[t].chi_f;
		for (int s=0; s<blocchi[t].quante_fibre(); s++)
			ris += (norma(blocchi[t].param[s][0], blocchi[t].param[s][1])+2);
	}
	ris += (6 * (1 - chi_tot));
	return ris;
}

//auxiliary complexity function, whose minimum on representations of the
//manifold gives c3
//(recall that the same graph manifold can be represented via parameters
//in many ways)
int graph_manifold::d3() const {
  int ris = 0;
  for (int t=0; t<A.size(); t++) {
    matrice B = J * A[t] * J;
		if (blocchi[t+1].quante_fibre() == 3) {
			matrice C(1,0,-1,1);
			if (norma(C * B) < norma(B))
				B = C * B;
		}
		if (!self_assembling() && blocchi[0].chi_f == 0)
			if (norma(B * J) < norma(B))    //sarebbe J * (J * B * J) = B * J
				B = B * J;
    ris+= norma(B);
		if (self_assembling() && norma(B)>0)
			ris--;
  }
  return d3_senza_matrici() + ris;
}

//lists all lens spaces of complexity c
int lens(int c) {
  if (c==0) return 3;
  if (c==1) return 2;
  int pw = 1;
  for (int t=0; t<c/2-1; t++)
    pw *=2;
  return ((c % 2) ? 2*lens(c-1) : 2*lens(c-1) - pw);
}

//se |p,q|<=n, allora |p|,|q|<=fib(n+1)
int fib(int i) {
  if (i<0) return 0;
  if (i==1 || i==0) return 1;
  return fib(i-2) + fib(i-1);
}

//lists all Seifert manifolds of complexity c, obtained with copies
//of B2, B3, B4, see [1]
void Seifert_con_B4(Seifert_manifold& M, int liv, int c, int& quanti, std::vector<int>& quanti_geo) {
  static int quanti_inizio;
  if (liv == 0) quanti_inizio = quanti;
  int num_fibre = M.param.size();
  if (liv < num_fibre) {
    int c_parziale = - 6*(M.chi_f-1) + 2*num_fibre;
    for (int t=0; t<liv; t++)
      c_parziale += M.norma(t);
    if (c_parziale > c) return;
    int max_param = fib(c - c_parziale + 1);
    for (M.param[liv][0] = liv ? M.param[liv-1][0] : 2; M.param[liv][0]<=max_param; M.param[liv][0]++)
      for (M.param[liv][1] = (liv && M.param[liv-1][0] == M.param[liv][0]) ? M.param[liv-1][1] : 1; 
					 M.param[liv][1] < M.param[liv][0]; M.param[liv][1]++)
				if (::MCD(M.param[liv][0], M.param[liv][1]) == 1 && M.norma(liv) < c)
					Seifert_con_B4(M, liv+1, c, quanti, quanti_geo);
  }
  else {
    for (M.b = -(num_fibre/2); M.b<10; M.b++) {
      if (M.forma_standard() && 
					c==M.d3() && !M.cij() && !M.from_B5()) {
				int fatti = quanti - quanti_inizio;
				if (fatti > 0 && !(fatti % Seifert_righe_per_pagina)) {
					std::cout << "\\end{tabular} \\end{center}" << std::endl;
					std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c}" << std::endl;
					std::cout << "N & fibers & $b$ & geometry \\\\ \\hline" << std::endl;
				}
        M.scrivi(++quanti);
        quanti_geo[M.geometria()]++;
      }
    }
  }
}	

//lists all torus fibers of complexity c (c is length(v)-5)
void torus_fibers(int liv, std::vector<BOOL>& v, int& quanti, std::vector<int>& quanti_geo) {
  if (liv < v.size()) {
    for (int i=0; i<=1; i++) {
      v[liv] = i;
      torus_fibers(liv+1, v, quanti, quanti_geo);
    }
  }
  else {
    BOOL minima = TRUE;
    for (int t=0; t < v.size(); t++) {
      std::vector<BOOL> v2 = v;
      rotate(v2.begin(), v2.begin()+t, v2.end());
      if (v2<v) minima = FALSE;
      for (int i=0; i<v2.size(); i++) v2[i] = 1-v2[i];
      if (v2<v) minima = FALSE;
    }
    if (minima && count(v.begin(), v.end(), 1)) {
      quanti_geo[SOL]+=2;
      matrice A = I;
      matrice D(1,0,1,1);
      matrice U(1,1,0,1);
      for (int i=0; i<v.size(); i++)
				A = v[i] ? D*A : U*A;
      if (output_in_tex) {
				std::cout << "$$";
				for (int sign=0; sign<=1; sign++) {
					std::cout << ++quanti << ": T_{\\matr";
					for (int t=0; t<=1; t++)
						for (int u=0; u<=1; u++)
							std::cout << "{" << A[t][u] << "}";
					std::cout << "}";
					std::cout << (sign ? "$$ " : " \\quad ");
					A = -A;
				}
      }
      else for (int sign=0; sign<=1; sign++) {
				std::cout << std::setw(4) << ++quanti << ": ";
				for (int t=0; t < v.size(); t++) std::cout << v[t];
				std::cout << (sign ? " (-)" : "    ");
				std::cout << "                    Sol\n";
      }
    }
  }
}

//lists graph manifolds of complexity c of 'type' M
//(i.e., with a fixed structure of blocks, given by M: only fibers and matrices vary)
//(works only for the configurations one find for c<12)
//(in case of self-assemblings, does not guarantee that manifolds are all distinct)
void grafo(int liv1, int liv2, graph_manifold& M, int c, int& quanti, std::vector<int>& quanti_geo,
					 BOOL cerca_Sol) {
  static int quanti_inizio;
  if (liv1 == 0 && liv2 == 0)
		quanti_inizio = quanti;
  if (liv1<M.blocchi.size()) {
    int max_param = fib(c - M.d3_senza_matrici() + 2);  //puo' essere migliorato?
    if (cerca_Sol) max_param = 2;
    for (int p = 2; p<=max_param; p++)
      for (int q = 1; q<p; q++)
				if (::MCD(p,q) == 1) {
					M.blocchi[liv1].param[liv2][0] = p;
					M.blocchi[liv1].param[liv2][1] = q;
          BOOL ok = TRUE;
					
					//checks that the parameters in the block are ordered
          if (liv2 >= 1)
            for (int t=0; t<liv2; t++)
              if (M.blocchi[liv1].param[liv2] < M.blocchi[liv1].param[t]) ok = FALSE;
					
					//checks whether it is the last fiber of the block
          BOOL ultimo2 = (liv2 == M.blocchi[liv1].quante_fibre()-1);
          if (ok) grafo(ultimo2 ? liv1+1 : liv1, ultimo2 ? 0 : liv2+1, 
												M, c, quanti, quanti_geo, cerca_Sol);
					
				}
    //ripristina il (2,1), altrimenti sbaglia sopra a contare max_param
    M.blocchi[liv1].param[liv2][0] = 2;
    M.blocchi[liv1].param[liv2][1] = 1;
  }
  else {    
    int somma_c_matrici = c - M.d3_senza_matrici();
    if (somma_c_matrici >= 0) {
      
      //il vettore v contiene numeri in 0,1,2, servira' per scrivere
      //la matrice come epsilon*S[v[0]]*J*S[v[1]]*...*J*S[v[l-1]]*S[1]^m con m=1 se det=-1
      int l = somma_c_matrici + M.A.size();
			
			//se M e' un autoassemblaggio di un pezzo:
			//se somma_c_matrici > 0 allora somma_c_matrici+=1,
			//se somma_c_matrici = 0 vanno svolti i casi somma_c_matrici = 0 e 1 entrambi
			for (int caso_self = 0; caso_self==0 || (!somma_c_matrici && M.self_assembling()
																							 && caso_self<=1); caso_self++) {
				if (M.self_assembling())
					l+= (somma_c_matrici == 0 ? caso_self : 1);
				std::vector<int> v(l,0);
				std::vector<int> fine(l,0);
				
				do {
					
					//se M.A.size()==2 spezza il vettore v in due pezzi e li usa per definire M.A[0] e M.A[1]
					for (int t=1; t==1 || (M.A.size() == 2 && t<l); t++) {
						
						//controlla che v[i]=2 al piu' solo agli estremi SERVE ?
						BOOL due_solo_estremi = TRUE;
						for (int i=1; i<l-1; i++)
							if (v[i] == 2 && !(M.A.size()==2 && (i==t || i==t-1))) 
								due_solo_estremi = FALSE;
						if (due_solo_estremi) {
							graph_manifold N = M;
							
							//at first, uses coordinates (-m,l)
							for (int sign_det = 0; sign_det<2*M.A.size(); sign_det++) {
								N.A[0] = S[v[0]];
								if (N.A.size() == 1) 
									for (int h=1; h<l; h++)
										N.A[0] = N.A[0] * J * S[v[h]];
								else {
									for (int h=1; h<t; h++)
										N.A[0] = N.A[0] * J * S[v[h]];
									N.A[1] = S[v[t]];
									for (int h=t+1; h<l; h++)
										N.A[1] = N.A[1] * J * S[v[h]];
								}
								if (sign_det % 2) N.A[0] = N.A[0] * S[1];
								if (sign_det > 1) N.A[1] = N.A[1] * S[1];
								
								//if the second block has 3 fibers, one can also create
								//|p   -q  |                |p q|
								//|-r-p s-q|   with norm of |r s| equals to l
								//if the first block is a mobius strip with one fiber, one can also create
								//|p   q |                |p q|
								//|-r  -s|   with norm of |r s| equals to l
								for (int caso = 0; caso == 0 || (caso <= 1 && N.blocchi.size()>1 && 
																								 (N.blocchi[1].quante_fibre()==3 ||
																									N.blocchi[0].chi_f==0)); caso++) {
									
									//now changes coordinates back, and A changes from
									//|p q|      |p -q|
									//|r s|   to |-r s|
									if (caso == 0) for (int h=0; h<N.A.size(); h++)
										N.A[h] = J * N.A[h] * J;
									
									//if the first block is a Moebius strip, changes A from
									//|p q|      |-p -q|
									//|r s|   to |r   s|
									if (caso == 1 && N.blocchi[0].chi_f == 0) {
										N.A[0] = J * N.A[0] * J;
										N.A[0] = J * N.A[0];
									}
									
									//if the second block has 3 fibers, another matrix can be created, see above
									if (caso == 1 && N.blocchi[1].quante_fibre() == 3) {
										matrice B(1,0,-1,1);
										N.A[0] = B * N.A[0];
									}
									
									//change signs if necessary
									for (int h=0; h<N.A.size(); h++)
										if (-N.A[h] < N.A[h]) N.A[h] = -N.A[h];
									
									//if there is a selfassemblig, signs are actually important
									for (int sign = 0; sign==0 || (N.self_assembling() && sign<=1); sign++) {
										if (sign) 
											N.A[0] = -N.A[0];
										
										//check it is not Seifert  
										if (N.orientable() && N.is_JSJ() && !cerca_Sol ||
												N.is_Sol() && cerca_Sol) {
											if (N.d3() == c && N.forma_standard()) {
												int fatti = quanti - quanti_inizio;
												if (output_in_tex && fatti > 0 && !(fatti % graph_righe_per_pagina)) {
													std::cout << "\\end{tabular} \\end{center}" << std::endl;
													if (N.blocchi.size()==2) {
														std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c}" << std::endl;
														std::cout << "N & fibers & matrix & fibers \\\\ \\hline" << std::endl;
													}
													if (N.blocchi.size()==3) {
														std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c|c|c}" << std::endl;
														std::cout << "N & fibers & matrix & fibers & matrix & fibers \\\\ \\hline" 
																			<< std::endl;
													}
												}
												N.scrivi(++quanti);
												quanti_geo[cerca_Sol ? SOL : GRAPH]++;
											}
										}
									}
								}
							}
						}
					}
					v[l-1] = (v[l-1] + 1) % 3;
					for (int t=l-1; t>=1; t--)
						if (v[t]) break;
						else v[t-1] = (v[t-1] + 1) % 3;
				} while (v != fine);
			}
		}
	}
}

//main !
int main() {
  int c;
  if (!output_in_tex) std::cout << "Inserire complessita': ";
  std::cin >> c;
	
  if (output_in_tex) {
    std::cout << "\\documentclass [11pt,oneside]{article}" << std::endl;
    std::cout << "\\reversemarginpar \\textwidth=14cm \\textheight=19cm" << std::endl;
    std::cout << "\\usepackage{amssymb} \\usepackage{amsfonts} \\usepackage{amsmath}" << std::endl;
    std::cout << "\\newcommand{\\matr} [4] {\\phantom{\\Big|}\\tiny{\\left(\\begin{array} {@{}c@{\\ }c@{}} #1 & #2";
    std::cout << "\\\\ #3 & #4 \\\\ \\end{array} \\right)}}" << std::endl;
    std::cout << "\\begin{document}" << std::endl;
  }

  //lens spaces:
  int quanti = 0;
  std::vector<int> quanti_geo(9,0);
  quanti_geo[LENS] = lens(c);
  quanti = quanti_geo[LENS];
  if (output_in_tex) {
    std::cout << "\\paragraph{Lens spaces: }" << std::endl; 
    for (int p=2; p<=fib(c+3); p++)
      for (int q=1; q<=p/2; q++)
				if (MCD(p,q) == 1 && norma(p,q) == c+2) {
					BOOL gia_vista = FALSE;
					for (int q2 = 1; q2<q; q2++)
						if ((q*q2) % p == 1 || (q*q2) % p == p-1) gia_vista = TRUE;
					if (!gia_vista) std::cout << "$L_{" << p << ", " << q << "}$ \\ ";
				}
  }
  else {
    std::cout << "-----------\n";
    std::cout << "Lenticolari:  " << quanti_geo[LENS] << "\n";
    std::cout << "-----------\n";
  }

  //Seifert manifolds obtained with B2,B3,B4:
  if (c>3) for (int chi_f = 2; chi_f>=chi_min; chi_f--) if (c>5 || chi_f == 2)
    for (int orient = 0; orient<=1; orient++)
      if ((chi_f==2 && orient) || 
					(chi_f<2 && (!(chi_f % 2) || orient == 0))) {
				if (output_in_tex) {
					std::cout << "\\paragraph{Seifert spaces over ";
					if (chi_f == 2) std::cout << "$S^2$:";
					if (chi_f == 1) std::cout << "$\\mathbb{P}^2$:";
					if (chi_f == 0 && orient) std::cout << "the torus:";
					if (chi_f == 0 && !orient) std::cout << "the Klein bottle:";
					std::cout << "}" << std::endl;
					std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c}" << std::endl;
					std::cout << "N & fibers & $b$ & geometry \\\\ \\hline" << std::endl;
				}
				else {
					std::cout << "-----------\n";
					std::cout << "chi_f = " << chi_f << " ";
					std::cout << (orient ? "orientabile" : "non orientabile") << "\n";
					std::cout << "-----------\n";
				}
				int num_fibre_min = chi_f+1;
				if (chi_f<=0) num_fibre_min = 0;
				for (int num_fibre = num_fibre_min; num_fibre <= num_fibre_max; num_fibre++) {
					tensor2 param(num_fibre, std::vector<int>(2,0));
          Seifert_manifold M(chi_f, param);
					Seifert_con_B4(M, 0, c, quanti, quanti_geo);
				}
				if (output_in_tex)
					std::cout << "\\end{tabular} \\end{center}" << std::endl;
      }

  //C(i,j):
  tensor2 param(3, std::vector<int>(2,0));
  if (c>=2) {
    if (output_in_tex) {
      std::cout << "\\paragraph{$C_{i,j}$ (Seifert with base space $S^2$):}" << std::endl;
      std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c}" << std::endl;
      std::cout << "N & fibers & $b$ & geometry \\\\ \\hline" << std::endl;
    }
    else {
      std::cout << "-----------\n";
      std::cout << "Cij:\n";
      std::cout << "-----------\n";
    }
    param[0][0] = 2; param[0][1] = 1;
    param[1][1] = 1; param[2][1] = 1;
    Seifert_manifold M(2, param, -1);
    for (int i=1; i<c; i++) {
      int j=c-i;
      M.param[1][0] = 1+j; M.param[2][0] = 1+i;
      if (i>=j && !(j==2 && i>=4)) {
				M.scrivi(++quanti); 
				quanti_geo[M.geometria()]++;
      }			
    }
    if (output_in_tex) 
      std::cout << "\\end{tabular} \\end{center}" << std::endl;
  }

  //E(k):
  if (c>=5) {
    if (output_in_tex) {
      std::cout << "\\paragraph{$E_k$ (Seifert with base space $S^2$):}" << std::endl;
      std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c}" << std::endl;
      std::cout << "N & fibers & $b$ & geometry \\\\ \\hline" << std::endl;
    }
    else {
      std::cout << "-----------\n";
      std::cout << "Ek:\n";
      std::cout << "-----------\n";
    }
    param[0][0] = 2; param[0][1] = 1;
    param[1][0] = 3; param[1][1] = 1; 
    param[2][0] = c; param[2][1] = 1;
    Seifert_manifold M(2, param, -1);
    M.scrivi(++quanti);
    quanti_geo[M.geometria()]++;
    if (output_in_tex) 
      std::cout << "\\end{tabular} \\end{center}" << std::endl;
  }
	
  //Seifert obtained using B5:
  if (c>=8) {
    if (output_in_tex) {
      std::cout << "\\paragraph{Using $B_5$ (Seifert with base space $S^2$):}" << std::endl;
      std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c}" << std::endl;
      std::cout << "N & fibers & $b$ & geometry \\\\ \\hline" << std::endl;
    }
    else {
      std::cout << "-----------\n";
      std::cout << "Using B5:\n";
      std::cout << "-----------\n";
    }
    for (int p=7; p<=fib(c); p++)  //puo' essere migliorato
      for (int q=2; 5*q<p; q++)
				if (::MCD(p,q)==1 && c == norma(p,q)+2) {
					param[2][0] = p; param[2][1] = q;
					Seifert_manifold M(2, param, -1);
					M.scrivi(++quanti);
					quanti_geo[M.geometria()]++;
				}
    if (output_in_tex) 
      std::cout << "\\end{tabular} \\end{center}" << std::endl;
  }

  //Sol manifolds:
  if (c>=7) {
    if (output_in_tex) {
      std::cout << "\\paragraph{Sol:}" << std::endl;
    }
    else {
      std::cout << "-----------\n";
      std::cout << "Sol:\n";
      std::cout << "-----------\n";
    }

    //torus fiberings:
    std::vector<BOOL> v(c-5, 0);
    torus_fibers(0, v, quanti, quanti_geo);

    //fiberings over 1-orbifold:
    if (output_in_tex) {
      std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c}" << std::endl;
      std::cout << "N & fibers & matrix & fibers \\\\ \\hline" << std::endl;
    }
    tensor2 param(2, std::vector<int>(2,0));
    Seifert_manifold blocco(1, param);
    std::vector<matrice> A(1);
    tensor2 collegamenti(1, std::vector<int>(2,0));
    collegamenti[0][1] = 1;
    graph_manifold N(std::vector<Seifert_manifold>(2, blocco), collegamenti, A);
    grafo(0, 0, N, c, quanti, quanti_geo, 1);
    if (output_in_tex) 
      std::cout << "\\end{tabular} \\end{center}" << std::endl;

  }
	
  //graph manifolds:
  if (c>=7) {
    if (output_in_tex) {
      std::cout << "\\paragraph{Non-geometric graph manifolds:}" << std::endl;
      std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c}" << std::endl;
      std::cout << "N & fibers & matrix & fibers \\\\ \\hline" << std::endl;
    }
    else {
      std::cout << "-----------\n";
      std::cout << "Non-geometric graph manifold:\n";
      std::cout << "-----------\n";
    }
    
    std::vector<int> coppia(2);
    coppia[0] = 2; coppia[1] = 1;
    tensor2 param(2, coppia);
    Seifert_manifold blocco_disco(1, param);
    std::vector<matrice> A(1);
    tensor2 collegamenti(1, std::vector<int>(2,0));
    collegamenti[0][1] = 1;
    graph_manifold N(std::vector<Seifert_manifold>(2, blocco_disco), collegamenti, A);
    grafo(0, 0, N, c, quanti, quanti_geo, 0);

    if (output_in_tex) 
      std::cout << "\\end{tabular} \\end{center}" << std::endl;

    if (c>=9) {
      if (output_in_tex) {
				std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c|c|c}" << std::endl;
				std::cout << "N & fibers & matrix & fibers over $A$ & matrix & fibers \\\\ \\hline" << std::endl;
      }
      
      //disco con due fibre, anello con una fibra, disco con due fibre
      tensor2 param_anello(1, coppia);
      Seifert_manifold blocco_anello(0, param_anello);
      std::vector<Seifert_manifold> blocchi(1, blocco_disco);
      blocchi.push_back(blocco_anello);
      blocchi.push_back(blocco_disco);    
      std::vector<matrice> A2(2);
      tensor2 collegamenti2(2, std::vector<int>(2,0));
      collegamenti2[0][1] = 1; collegamenti2[1][0] = 1; collegamenti2[1][1] = 2;
      graph_manifold N2(blocchi, collegamenti2, A2);
      grafo(0, 0, N2, c, quanti, quanti_geo, 0);

      if (output_in_tex) {
				std::cout << "\\end{tabular} \\end{center}" << std::endl;
				std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c}" << std::endl;
				std::cout << "N & fibers & matrix & fibers \\\\ \\hline" << std::endl;
			}
      
      //disco con tre fibre, disco con due fibre
      tensor2 param_disco3(3, coppia);
      Seifert_manifold blocco_disco3(1, param_disco3);
      blocchi = std::vector<Seifert_manifold>(1, blocco_disco);
      blocchi.push_back(blocco_disco3);    
      graph_manifold N3(blocchi, collegamenti, A);
      grafo(0, 0, N3, c, quanti, quanti_geo, 0);
      if (output_in_tex) {
				std::cout << "\\end{tabular} \\end{center}" << std::endl;
				std::cout << "\\begin{center} \\begin{tabular}{c|c|c|c}" << std::endl;
				std::cout << "N & fibers over $S$ & matrix & fibers \\\\ \\hline" << std::endl;
			}

      //mobius con una fibra, disco con due fibre
			tensor2 param_mobius(1, coppia);
      Seifert_manifold blocco_mobius(0, param_mobius);
      blocchi = std::vector<Seifert_manifold>(1, blocco_mobius);
      blocchi.push_back(blocco_disco);    
      graph_manifold Nmob(blocchi, collegamenti, A);
      grafo(0, 0, Nmob, c, quanti, quanti_geo, 0);
      if (output_in_tex) {
				std::cout << "\\end{tabular} \\end{center}" << std::endl;
				std::cout << "\\begin{center} \\begin{tabular}{c|c|c}" << std::endl;
				std::cout << "N & fibers over $A$ & matrix \\\\ \\hline" << std::endl;
			}
			
			//anello con una fibra
			blocchi = std::vector<Seifert_manifold>(1, blocco_anello);	
			tensor2 collegamenti_self(1, std::vector<int>(2,0));
			graph_manifold Nself(blocchi, collegamenti_self, A);
      grafo(0, 0, Nself, c, quanti, quanti_geo, 0);			

      if (output_in_tex) 
				std::cout << "\\end{tabular} \\end{center}" << std::endl;
		}
  }
  
  //if c==6, a flat manifold has been counted twice
  if (c == 6) {
    quanti_geo[FLAT]--; quanti--;
  }

  //summarizing table:
  if (output_in_tex) {
    std::cout << "\\paragraph{Summarizing table:}" << std::endl;
    std::cout << "\\begin{center} \\begin{tabular}{|r|c|} \\hline ";
    for (int i=0; i<9; i++) 
      std::cout << nome_geometria_tex[i] << " & $" << quanti_geo[i] << "$ \\\\" << std::endl;
    std::cout << "\\hline total & $" << quanti << "$ \\\\ \\hline" << std::endl;
    std::cout << "\\end{tabular} \\end{center}";
  }
  else {
    for (int i=0; i<9; i++) 
      std::cout << std::setw(20) << nome_geometria[i] << ": " << quanti_geo[i] << "\n";
    std::cout << std::setw(20) << "Total" << ": " << quanti << "\n";
  }
  if (c == 6) std::cout << "A flat manifold has been counted twice.\n";
  if (output_in_tex) std::cout << "\\end{document}" << std::endl;



  return 0;
}
